from typing import Any, Callable, Optional, Sequence, Union

import equinox as eqx
import jax
import jax.random as jrandom


class Chain(eqx.Module):
    """Similar to `equinox.nn.Sequence` but also allows callables in the sequence."""

    layers: Sequence[Union[eqx.Module, Callable]]

    def __call__(self, x: Any, *, key: Optional["jax.random.PRNGKey"] = None) -> Any:
        """**Arguments:**

        - `x`: Argument passed to the first member of the sequence.
        - `key`: A `jax.random.PRNGKey`, which will be split and passed to every layer
            to provide any desired randomness. (Optional. Keyword only argument.)
        **Returns:**

        The output of the last member of the sequence.
        """

        if key is None:
            keys = [None] * len(self.layers)
        else:
            keys = jrandom.split(key, len(self.layers))
        for layer, key in zip(self.layers, keys):
            if isinstance(layer, eqx.Module):
                x = layer(x, key=key)
            else:
                x = layer(x)
        return x

    def __getitem__(self, i: Union[int, slice]) -> eqx.Module:
        if isinstance(i, int):
            return self.layers[i]
        elif isinstance(i, slice):
            return Chain(self.layers[i])
        else:
            raise TypeError(f"Indexing with type {type(i)} is not supported")


Chain.__init__.__doc__ = """**Arguments:**
- `layers`: A sequence of `equinox.Module`s and/or Callables.
"""



from .alexnet import AlexNet, alexnet
from .googlenet import GoogLeNet, googlenet
from .resnet import (
    ResNet,
    resnet18,
    resnet34,
    resnet50,
    resnet101,
    resnet152,
    resnext50_32x4d,
    resnext101_32x8d,
    wide_resnet50_2,
    wide_resnet101_2,
)
from .vit import (
    VisionTransformer,
    vit_base,
    vit_small,
    vit_tiny,
    VitAttention,
    VitBlock,
)



import equinox as eqx
import jax
import jax.random as jrandom
from equinox.custom_types import Array


class DropPath(eqx.Module):
    """Effectively dropping a sample from the call.
    Often used inside a network along side a residual connection."""

    p: float
    inference: bool

    def __init__(self, p: float = 0.0, inference: bool = False):
        """**Arguments:**

        - `p`: The probability to drop a sample entirely during forward pass.
        - `inference`: Defaults to `False`. If `True`, then the input is returned unchanged.
        This may be toggled with `equinox.tree_inference`.
        """
        self.p = p
        self.inference = inference

    def __call__(self, x, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: An any-dimensional JAX array to drop.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for calculating
            which elements to dropout. (Keyword only argument.)
        """
        if self.inference or self.p == 0.0:
            return x
        if key is None:
            raise RuntimeError(
                "DropPath requires a key when running in non-deterministic mode. Did you mean to enable inference?"
            )

        keep_prob = 1 - self.p
        return x * jrandom.bernoulli(key, p=keep_prob)
